# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.tektronix.tds2024b
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control an oscilloscope.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import struct

import numpy as np

from lantz.feat import Feat
from lantz.action import Action
from lantz.drivers.legacy.visa import VisaDriver

class TDS2024(VisaDriver):
    """Tektronix TDS2024 200 MHz 4 Channel Digital Real-Time Oscilloscope
    """

    def __init__(self, port):
        super().__init__(port)
        timeout=10

    @Action()
    def autoconf(self):
        """Autoconfig oscilloscope.
        """
        self.send(':AUTOS EXEC')

    def initialize(self):
        """initiate.
        """
        self.send(':ACQ:STATE ON')
        return "Init"

    @Feat()
    def idn(self):
        """IDN.
        """
        return self.query('ID?')

    @Feat()
    def trigger(self):
        """Trigger state.
        """
        return self.query(':TRIG:STATE?')

    @trigger.setter
    def trigger(self, mode):
        """Set trigger state.
        """
        self.query('TRIG:MAIN:MODE {}'.format(mode))

    @Action()
    def triggerlevel(self):
        """Set trigger level to 50% of the minimum adn maximum
        values of the signal.
        """
        self.send('TRIG:MAIn SATLevel')

    @Action()
    def forcetrigger(self):
        """Force trigger event.
        """
        self.send('TRIG FORCe')

    @Action()
    def datasource(self, chn):
        """Selects channel.
        """
        self.send(':DATA:SOURCE CH{}'.format(chn))

    @Action()
    def acqparams(self):
        """ X/Y Increment Origin and Offset.
        """
        commands = 'XZE?;XIN?;YZE?;YMU?;YOFF?'
        #params = self.query(":WFMPRE:XZE?;XIN?;YZE?;YMU?;YOFF?;")
        params = self.query(':WFMPRE:{}'.format(commands))
        params = {k: float(v) for k, v in zip(commands.split(';'), params.split(';'))}
        return params

    @Action()
    def dataencoding(self):
        """Set data encoding.
        """
        self.send(':DAT:ENC RPB;WID 2;')
        return "Set data encoding"

    @Action()
    def curv(self):
        """Get data.

            Returns:
            xdata, data as list
        """
        self.dataencoding()
        self.send('CURV?')
        answer = self.recv()
        numdigs = int(answer[1])
        bytecount = int(answer[2:2+numdigs])
        data = answer[2+numdigs:]
        length = bytecount / 2
        data = struct.unpack("{}H".format(length), data[0:2*length])
        params = self.acqparams()
        data = np.array(list(map(float, data)))
        yoff = params['YOFF?']
        ymu = params['YMU?']
        yze = params['YZE?']
        xin = params['XIN?']
        xze = params['XZE?']
        ydata = ( data - yoff) * ymu + yze
        xdata = np.arange(len(data)) * xin + xze
        return list(xdata), list(data)

    def _measure(self, type, source):
        self.send('MEASUrement:IMMed:TYPe {}'.format(type))
        self.send('MEASUrement:IMMed:SOUrce1 CH{}'.format(source))
        self.send('MEASUrement:IMMed:VALue?')
        return self.recv()

    @Action()
    def measure_frequency(self, channel):
        """Get immediate measurement result.
        """
        return self._measure('FREQuency', channel)

    @Action()
    def measure_min(self, channel):
        """Get immediate measurement result.
        """
        return self._measure('MINImum', channel)

    @Action()
    def measure_max(self, chn):
        """Get immediate measurement result.
        """
        return self._measure('MAXImum', channel)

    @Action()
    def measure_mean(self, chn):
        """Get immediate measurement result.
        """
        return self._measure('MEAN', channel)


if __name__ == '__main__':
    import argparse
    import csv

    parser = argparse.ArgumentParser(description='Measure using TDS2024 and dump to screen')
    parser.add_argument('-p', '--port', default='USB0::0x0699::0x036A::C048617',
                        help='USB port')
    parser.add_argument('-v', '--view', action='store_true', default=False,
                        help='View ')
    parser.add_argument('Channels', metavar='channels', type=int, nargs='*',
                        help='Channels to use')
    parser.add_argument('--output', type=argparse.FileType('wb', 0), default='-')

    args = parser.parse_args()

    osc = TDS2024(args.port)
    osc.initialize()
    print(osc.idn)
    print(osc.trigger)
    osc.forcetrigger()
    osc.triggerlevel()
    osc.trigger = "AUTO"
    print(osc.trigger)
    #osc.autoconf()
    params = osc.acqparams()

    if args.view:
        import matplotlib.pyplot as plt
        import numpy as np

    with args.output as fp:
        writer = csv.writer(fp)
        writer.write_row(('Channel', 'Freq', 'Max', 'Min', 'Mean'))
        for channel in args.channels or range(1, 4):
            osc.datasource(channel)
            writer.write_row(([osc.measure_frequency(channel),
                               osc.measure_max(channel),
                               osc.measure_min(channel),
                               osc.measure_mean(channel)]))

            if args.view:
                x, y = osc.curv()
                x = np.array(x)
                x = x - x.min()
                y = np.array(y)
                plt.plot(x, y)

    if args.view:
        plt.show()



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.tektronix.tds1002b
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control an oscilloscope.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.

    Source: Tektronix Manual
"""

from lantz import Feat
from lantz.drivers.usbtmc import USBTMCDriver


class TDS1002b(USBTMCDriver):

    @classmethod
    def usb_from_serial(cls, serial_number):
        # find resource and get resource name

        #super().__init__(1689, 867, serial_number, **kwargs)

        return cls(resource_name='bla')

    @Feat(read_once=True)
    def idn(self):
        return self.query('*IDN?')



# -*- coding: utf-8 -*-
"""
    lantz.drivers.basler
    ~~~~~~~~~~~~~~~~~~~~~~
    :company: Basler
    :description: Manufactures cameras industrial applications
    :website: https://baslerweb.com/en
    ----
    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from .basler import BaslerCam

__all__ = ['BaslerCam']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.ni.daqmx.tasks
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of specialized tasks clases.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np

from lantz import Feat, Action
from lantz.foreign import RetStr, RetTuple, RetValue

from .base import Task, Channel
from .constants import Constants

_GROUP_BY = {'scan': Constants.Val_GroupByScanNumber,
             'channel': Constants.Val_GroupByChannel}



class AnalogInputTask(Task):
    """Analog Input Task
    """

    IO_TYPE = 'AI'

    @Feat()
    def max_convert_rate(self):
        """Maximum convert rate supported by the task, given the current
        devices and channel count.

        This rate is generally faster than the default AI Convert
        Clock rate selected by NI-DAQmx, because NI-DAQmx adds in an
        additional 10 microseconds per channel settling time to
        compensate for most potential system settling constraints.

        For single channel tasks, the maximum AI Convert Clock rate is the
        maximum rate of the ADC. For multiple channel tasks, the maximum
        AI Convert Clock rate is the maximum convert rate of the analog
        hardware, including the ADC, filters, multiplexers, and amplifiers.
        Signal conditioning accessories can further constrain the maximum AI
        Convert Clock based on timing and settling requirements.
        """
        err, value = self.lib.GetAIConvMaxRate(RetValue('f64'))
        return value

    def read_scalar(self, timeout=10.0):
        """Return a single floating-point sample from a task that
        contains a single analog input channel.

        :param timeout:  The amount of time, in seconds, to wait for the function to
                         read the sample(s). The default value is 10.0 seconds. To
                         specify an infinite wait, pass -1 (DAQmx_Val_WaitInfinitely).
                         This function returns an error if the timeout elapses.

                         A value of 0 indicates to try once to read the requested
                         samples. If all the requested samples are read, the function
                         is successful. Otherwise, the function returns a timeout error
                         and returns the samples that were actually read.

        :rtype: float
        """

        err, value = self.lib.ReadAnalogScalarF64(timeout, RetValue('f64'), None)
        return value

    @Action(units=(None, 'seconds', None), values=(None, None, _GROUP_BY))
    def read(self, samples_per_channel=None, timeout=10.0, group_by='channel'):
        """Reads multiple floating-point samples from a task that
        contains one or more analog input channels.

        :param samples_per_channel:
          The number of samples, per channel, to read. The default
          value of -1 (DAQmx_Val_Auto) reads all available samples. If
          readArray does not contain enough space, this function
          returns as many samples as fit in readArray.

          NI-DAQmx determines how many samples to read based on
          whether the task acquires samples continuously or acquires a
          finite number of samples.

          If the task acquires samples continuously and you set this
          parameter to -1, this function reads all the samples
          currently available in the buffer.

          If the task acquires a finite number of samples and you set
          this parameter to -1, the function waits for the task to
          acquire all requested samples, then reads those samples. If
          you set the Read All Available Samples property to TRUE, the
          function reads the samples currently available in the buffer
          and does not wait for the task to acquire all requested
          samples.

        :param timeout: float
          The amount of time, in seconds, to wait for the function to
          read the sample(s). The default value is 10.0 seconds. To
          specify an infinite wait, pass -1
          (DAQmx_Val_WaitInfinitely). This function returns an error
          if the timeout elapses.

          A value of 0 indicates to try once to read the requested
          samples. If all the requested samples are read, the function
          is successful. Otherwise, the function returns a timeout
          error and returns the samples that were actually read.

        :param group_by:

            'channel'
              Group by channel (non-interleaved)::

                ch0:s1, ch0:s2, ..., ch1:s1, ch1:s2,..., ch2:s1,..

            'scan'
              Group by scan number (interleaved)::

                ch0:s1, ch1:s1, ch2:s1, ch0:s2, ch1:s2, ch2:s2,...

        :rtype: numpy.ndarray
        """

        if samples_per_channel is None:
            samples_per_channel = self.samples_per_channel_available()

        number_of_channels = self.number_of_channels()
        if group_by == Constants.Val_GroupByScanNumber:
            data = np.zeros((samples_per_channel, number_of_channels), dtype=np.float64)
        else:
            data = np.zeros((number_of_channels, samples_per_channel), dtype=np.float64)

        err, data, count = self.lib.ReadAnalogF64(samples_per_channel, timeout, group_by,
                                                  data.ctypes.data, data.size, RetValue('i32'), None)

        if samples_per_channel < count:
            if group_by == 'scan':
                return data[:count]
            else:
                return data[:,:count]

        return data


class AnalogOutputTask(Task):
    """Analog Output Task
    """

    CHANNEL_TYPE = 'AO'

    @Action(units=(None, None, 'seconds', None), values=(None, None, None, _GROUP_BY))
    def write(self, data, auto_start=True, timeout=10.0, group_by='scan'):
        """Write multiple floating-point samples or a scalar to a task
        that contains one or more analog output channels.

        Note: If you configured timing for your task, your write is
        considered a buffered write. Buffered writes require a minimum
        buffer size of 2 samples. If you do not configure the buffer
        size using DAQmxCfgOutputBuffer, NI-DAQmx automatically
        configures the buffer when you configure sample timing. If you
        attempt to write one sample for a buffered write without
        configuring the buffer, you will receive an error.

        :param data: The array of 64-bit samples to write to the task
                     or a scalar.

        :param auto_start: Whether or not this function automatically starts
                           the task if you do not start it.

        :param timeout: The amount of time, in seconds, to wait for this
                        function to write all the samples. The default value is 10.0
                        seconds. To specify an infinite wait, pass -1
                        (DAQmx_Val_WaitInfinitely). This function returns an error
                        if the timeout elapses.

                        A value of 0 indicates to try once to write the submitted
                        samples. If this function successfully writes all submitted
                        samples, it does not return an error. Otherwise, the
                        function returns a timeout error and returns the number of
                        samples actually written.

        :param group_by: how the samples are arranged, either interleaved or noninterleaved

            'channel' - Group by channel (non-interleaved).

            'scan' - Group by scan number (interleaved).

        :return: The actual number of samples per channel successfully
                 written to the buffer.

        """
        if np.isscalar(data):
            err = self.lib.WriteAnalogScalarF64(bool32(auto_start),
                                                float64(timeout),
                                                float64(data), None)
            return 1

        data = np.asarray(data, dtype = np.float64)

        number_of_channels = self.number_of_channels()

        if data.ndims == 1:
            if number_of_channels == 1:
                samples_per_channel = data.shape[0]
                shape = (samples_per_channel, 1)
            else:
                samples_per_channel = data.size / number_of_channels
                shape = (samples_per_channel, number_of_channels)

            if not group_by == Constants.Val_GroupByScanNumber:
                shape = tuple(reversed(shape))

            data.reshape(shape)
        else:
            if group_by == Constants.Val_GroupByScanNumber:
                samples_per_channel = data.shape[0]
            else:
                samples_per_channel = data.shape[-1]

        err, count = self.lib.WriteAnalogF64(samples_per_channel, auto_start,
                                             timeout, group_by,
                                             data.ctypes.data, RetValue('i32'),
                                             None)

        return count


class DigitalTask(Task):

    @Action(units=(None, 'seconds', None), values=(None, None, _GROUP_BY))
    def read(self, samples_per_channel=None, timeout=10.0, group_by='scan'):
        """
        Reads multiple samples from each digital line in a task. Each
        line in a channel gets one byte per sample.

        :param samples_per_channel: int or None

          The number of samples, per channel, to
          read. The default value of -1 (DAQmx_Val_Auto) reads all
          available samples. If readArray does not contain enough
          space, this function returns as many samples as fit in
          readArray.

          NI-DAQmx determines how many samples to read based on
          whether the task acquires samples continuously or acquires a
          finite number of samples.

          If the task acquires samples continuously and you set this
          parameter to -1, this function reads all the samples
          currently available in the buffer.

          If the task acquires a finite number of samples and you set
          this parameter to -1, the function waits for the task to
          acquire all requested samples, then reads those samples. If
          you set the Read All Available Data property to TRUE, the
          function reads the samples currently available in the buffer
          and does not wait for the task to acquire all requested
          samples.

        :param timeout: float

          The amount of time, in seconds, to wait for the function to
          read the sample(s). The default value is 10.0 seconds. To
          specify an infinite wait, pass -1
          (DAQmx_Val_WaitInfinitely). This function returns an error
          if the timeout elapses.

          A value of 0 indicates to try once to read the requested
          samples. If all the requested samples are read, the function
          is successful. Otherwise, the function returns a timeout
          error and returns the samples that were actually read.

        :param group_by: {'group', 'scan'}

          Specifies whether or not the samples are interleaved:

            'channel' - Group by channel (non-interleaved).

            'scan' - Group by scan number (interleaved).

        Returns
        -------

          data : array

            The array to read samples into. Each `bytes_per_sample`
            corresponds to one sample per channel, with each element
            in that grouping corresponding to a line in that channel,
            up to the number of lines contained in the channel.

          bytes_per_sample : int

            The number of elements in returned `data` that constitutes
            a sample per channel. For each sample per channel,
            `bytes_per_sample` is the number of bytes that channel
            consists of.

        """

        if samples_per_channel in (None, -1):
            samples_per_channel = self.samples_per_channel_available()

        if self.one_channel_for_all_lines:
            nof_lines = []
            for channel in self.names_of_channels():
                nof_lines.append(self.number_of_lines (channel))
            c = int (max (nof_lines))
            dtype = getattr(np, 'uint%s' % (8 * c))
        else:
            c = 1
            dtype = np.uint8

        number_of_channels = self.number_of_channels()

        if group_by == Constants.Val_GroupByScanNumber:
            data = np.zeros((samples_per_channel, number_of_channels),dtype=dtype)
        else:
            data = np.zeros((number_of_channels, samples_per_channel),dtype=dtype)

        err, count, bps = self.lib.ReadDigitalLines(samples_per_channel, float64 (timeout),
              group_by, data.ctypes.data, uInt32 (data.size * c),
              RetValue('i32'), RetValue('i32'),
              None
        )
        if count < samples_per_channel:
            if group_by == 'scan':
                return data[:count], bps
            else:
                return data[:,:count], bps
        return data, bps


class DigitalInputTask(DigitalTask):
    """Exposes NI-DAQmx digital input task to Python.
    """

    CHANNEL_TYPE = 'DI'


class DigitalOutputTask(DigitalTask):
    """Exposes NI-DAQmx digital output task to Python.
    """

    CHANNEL_TYPE = 'DO'

    @Action(units=(None, None, 'seconds', None), values=(None, {True, False}, None, _GROUP_BY))
    def write(self, data, auto_start=True, timeout=10.0, group_by='scan'):
        """
        Writes multiple samples to each digital line in a task. When
        you create your write array, each sample per channel must
        contain the number of bytes returned by the
        DAQmx_Read_DigitalLines_BytesPerChan property.

	Note: If you configured timing for your task, your write is
	considered a buffered write. Buffered writes require a minimum
	buffer size of 2 samples. If you do not configure the buffer
	size using DAQmxCfgOutputBuffer, NI-DAQmx automatically
	configures the buffer when you configure sample timing. If you
	attempt to write one sample for a buffered write without
	configuring the buffer, you will receive an error.

        Parameters
        ----------

        data : array

          The samples to write to the task.

        auto_start : bool

          Specifies whether or not this function automatically starts
          the task if you do not start it.

        timeout : float

          The amount of time, in seconds, to wait for this function to
          write all the samples. The default value is 10.0 seconds. To
          specify an infinite wait, pass -1
          (DAQmx_Val_WaitInfinitely). This function returns an error
          if the timeout elapses.

          A value of 0 indicates to try once to write the submitted
          samples. If this function successfully writes all submitted
          samples, it does not return an error. Otherwise, the
          function returns a timeout error and returns the number of
          samples actually written.

        layout : {'group_by_channel', 'group_by_scan_number'}

          Specifies how the samples are arranged, either interleaved
          or noninterleaved:

            'group_by_channel' - Group by channel (non-interleaved).

            'group_by_scan_number' - Group by scan number (interleaved).
        """

        number_of_channels = self.get_number_of_channels()

        if np.isscalar(data):
            data = np.array([data]*number_of_channels, dtype = np.uint8)
        else:
            data = np.asarray(data, dtype = np.uint8)

        if data.ndims == 1:
            if number_of_channels == 1:
                samples_per_channel = data.shape[0]
                shape = (samples_per_channel, 1)
            else:
                samples_per_channel = data.size / number_of_channels
                shape = (samples_per_channel, number_of_channels)

            if not group_by == Constants.Val_GroupByScanNumber:
                shape = tuple(reversed(shape))

            data.reshape(shape)
        else:
            if group_by == Constants.Val_GroupByScanNumber:
                samples_per_channel = data.shape[0]
            else:
                samples_per_channel = data.shape[-1]

        err, count = self.lib.WriteDigitalLines(samples_per_channel,
                                                bool32(auto_start),
                                                float64(timeout), group_by,
                                                data.ctypes.data, RetValue('u32'), None)

        return count

        # NotImplemented: WriteDigitalU8, WriteDigitalU16, WriteDigitalU32, WriteDigitalScalarU32

class CounterInputTask(Task):
    """Exposes NI-DAQmx counter input task to Python.
    """

    CHANNEL_TYPE = 'CI'

    def read_scalar(self, timeout=10.0):
        """Read a single floating-point sample from a counter task. Use
        this function when the counter sample is scaled to a
        floating-point value, such as for frequency and period
        measurement.

        :param float:
          The amount of time, in seconds, to wait for the function to
          read the sample(s). The default value is 10.0 seconds. To
          specify an infinite wait, pass -1
          (DAQmx_Val_WaitInfinitely). This function returns an error if
          the timeout elapses.

          A value of 0 indicates to try once to read the requested
          samples. If all the requested samples are read, the function
          is successful. Otherwise, the function returns a timeout error
          and returns the samples that were actually read.

        :return: The sample read from the task.
        """

        err, value = self.lib.ReadCounterScalarF64(timeout, RetValue('f64'), None)
        return value

    def read(self, samples_per_channel=None, timeout=10.0):
        """Read multiple 32-bit integer samples from a counter task.
        Use this function when counter samples are returned unscaled,
        such as for edge counting.

        :param samples_per_channel:
          The number of samples, per channel, to read. The default
          value of -1 (DAQmx_Val_Auto) reads all available samples. If
          readArray does not contain enough space, this function
          returns as many samples as fit in readArray.

          NI-DAQmx determines how many samples to read based on
          whether the task acquires samples continuously or acquires a
          finite number of samples.

          If the task acquires samples continuously and you set this
          parameter to -1, this function reads all the samples
          currently available in the buffer.

          If the task acquires a finite number of samples and you set
          this parameter to -1, the function waits for the task to
          acquire all requested samples, then reads those samples. If
          you set the Read All Available Samples property to TRUE, the
          function reads the samples currently available in the buffer
          and does not wait for the task to acquire all requested
          samples.

        :param timeout:
          The amount of time, in seconds, to wait for the function to
          read the sample(s). The default value is 10.0 seconds. To
          specify an infinite wait, pass -1
          (DAQmx_Val_WaitInfinitely). This function returns an error
          if the timeout elapses.

          A value of 0 indicates to try once to read the requested
          samples. If all the requested samples are read, the function
          is successful. Otherwise, the function returns a timeout
          error and returns the samples that were actually read.


        :return: The array of samples read.
        """

        if samples_per_channel is None:
            samples_per_channel = self.samples_per_channel_available()

        data = np.zeros((samples_per_channel,),dtype=np.int32)

        err, count = self.lib.ReadCounterU32(samples_per_channel, float64(timeout),
                                             data.ctypes.data, data.size, RetValue('i32'), None)

        return data[:count]


class CounterOutputTask(Task):

    """Exposes NI-DAQmx counter output task to Python.
    """

    channel_type = 'CO'


Task.register_class(AnalogInputTask)



#Sacher_playground.py
"""
Make sure you're in the lantz32 environment to run this code properly
"""

import visa
import types
import ctypes
import sys
import numpy as np
import logging
from ctypes.wintypes import DWORD
from ctypes.wintypes import WORD
import ctypes.wintypes
import time

eposlib = ctypes.windll.LoadLibrary('C:\\Users\\Carbro\\Desktop\\Charmander\\EposCmd.dll')

DeviceName = b'EPOS'
ProtocolStackName = b'MAXON_RS232'
InterfaceName = b'RS232'


"""
    ok before I dive into this giant Sacher class thing let me just list here all the functions that are being defined in this class:

    These are the important ones:
    1) __init__(self, name, address, reset=False)
    2) open(self)
    3) initialize(self)
    4) get_offset(self)
    5) get_motor_position(self)
    6) do_get_wavelength(self)
    7) set_new_offset(self, new_offset)
    8) set_coeffs(self, a, b, c, min_wl, max_wl)
    9) do_set_wavelength(self, wavelength)
    10) set_target_position(self, target, absolute, immediately)

    And then there are these random ones:
    11) get_bit(self, byteval,idx)
    12) _u32todouble(self, uinput)
    13) _doubletou32(dinput)
    14) __del__(self)
    15) close(self)
    16) get_motor_current(self)
    17) find_home(self)
    18) restore(self)
    19) fine_tuning_steps(self, steps)
    20) is_open(self)
    21) clear_fault(self)

    They're labeled as #1, #2, ..., so you can find them with some ctrl+f
    """


class Sacher_EPOS():

#1)
    def __init__(self, name, address, reset=False):
        self._port_name = address
        self._is_open = False
        self._HPM = True
        self.open()
        self.initialize()
    """
    This is the special "__init__" function that gets runs automatically
    """

#2)
    def open(self):
        eposlib.VCS_OpenDevice.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.POINTER(DWORD)]
        eposlib.VCS_OpenDevice.restype = ctypes.wintypes.HANDLE
        buf = ctypes.pointer(DWORD(0))
        ret = ctypes.wintypes.HANDLE()
        ret = eposlib.VCS_OpenDevice(DeviceName, ProtocolStackName, InterfaceName, self._port_name, buf)
        self._keyhandle = ret
        if int(buf.contents.value) >= 0:
            self._is_open = True
            self._keyhandle = ret
        return

#3)
    """
    This uses u32todouble three times, for the three coefficients

    Okay here's the important part:
    self._coefA = CastedObjectData[0]

    self._doubleA = self._u32todouble(self._coefA)
    self._doubleB = self._u32todouble(self._coefB)
    self._doubleC = self._u32todouble(self._coefC)
    self._currentwl = self._doubleA*(self._offset)**2.0 + self._doubleB*self._offset + self._doubleC

    """
    def initialize(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        BaudRate = DWORD(38400)
        Timeout = DWORD(100)
        ret = eposlib.VCS_SetProtocolStackSettings(self._keyhandle,BaudRate,Timeout,ctypes.byref(buf))
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            raise ValueError(errbuf.value)
        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_ClearFault(self._keyhandle,nodeID,ctypes.byref(buf))
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            eposlib.VCS_GetErrorInfo(buf, errbuf, WORD(64))
            raise ValueError(errbuf.value)
        buf = ctypes.wintypes.DWORD(0)
        plsenabled = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_GetEnableState(self._keyhandle,nodeID,ctypes.byref(plsenabled),ctypes.byref(buf))
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            eposlib.VCS_GetErrorInfo(buf, errbuf, WORD(64))
            raise ValueError(errbuf.value)
        if int(plsenabled.value) != 0:
            logging.warning(__name__ + ' EPOS motor enabled, disabling before proceeding.')
            ret = eposlib.VCS_SetDisableState(self._keyhandle,nodeID,ctypes.byref(buf))
            if int(ret) != 0:
                logging.warning(__name__ + ' EPOS motor successfully disabled, proceeding')
            else:
                logging.error(__name__ + ' EPOS motor was not successfully disabled!')
        buf = ctypes.wintypes.DWORD(0)
        Counts = WORD(512) # incremental encoder counts in pulses per turn
        PositionSensorType = WORD(4)
        ret = eposlib.VCS_SetEncoderParameter(self._keyhandle,nodeID,Counts,PositionSensorType,ctypes.byref(buf))

        # Get operation mode, check if it's 1 -- this is "profile position mode"
        buf = ctypes.wintypes.DWORD(0)
        pMode = ctypes.pointer(ctypes.c_int8())
        eposlib.VCS_GetOperationMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.c_int8), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetOperationMode.restype = ctypes.wintypes.BOOL
        ret = eposlib.VCS_GetOperationMode(self._keyhandle, nodeID, pMode, ctypes.byref(buf))
        # if mode is not 1, make it 1
        if pMode.contents.value != 1:
            eposlib.VCS_SetOperationMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.c_int8, ctypes.POINTER(ctypes.wintypes.DWORD)]
            eposlib.VCS_SetOperationMode.restype = ctypes.wintypes.BOOL
            pMode_setting = ctypes.c_int8(1)
            ret = eposlib.VCS_SetOperationMode(self._keyhandle, nodeID, pMode_setting, ctypes.byref(buf))
        eposlib.VCS_GetPositionProfile.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetPositionProfile.restype = ctypes.wintypes.BOOL
        pProfileVelocity = ctypes.pointer(ctypes.wintypes.DWORD())
        pProfileAcceleration = ctypes.pointer(ctypes.wintypes.DWORD())
        pProfileDeceleration = ctypes.pointer(ctypes.wintypes.DWORD())
        ret = eposlib.VCS_GetPositionProfile(self._keyhandle, nodeID, pProfileVelocity, pProfileAcceleration, pProfileDeceleration,ctypes.byref(buf))

        #print(pProfileVelocity.contents.value, pProfileAcceleration.contents.value, pProfileDeceleration.contents.value)

        if (int(pProfileVelocity.contents.value) > int(11400) or int(pProfileAcceleration.contents.value) > int(60000) or int(pProfileDeceleration.contents.value) > int(60000)):
            eposlib.VCS_GetPositionProfile.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.DWORD, ctypes.wintypes.DWORD, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD)]
            eposlib.VCS_GetPositionProfile.restype = ctypes.wintypes.BOOL
            pProfileVelocity = ctypes.wintypes.DWORD(429)
            pProfileAcceleration = ctypes.wintypes.DWORD(429)
            pProfileDeceleration = ctypes.wintypes.DWORD(429)
            logging.warning(__name__ + ' GetPositionProfile out of bounds, resetting...')
            ret = eposlib.VCS_SetPositionProfile(self._keyhandle, nodeID, pProfileVelocity, pProfileAcceleration, pProfileDeceleration,ctypes.byref(buf))

        self._offset = self.get_offset()

        """DC - These are hardcoded values I got from the LabVIEW program -- I don't think any documentation exists on particular object indices"""

        """Coefficient A"""
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(1)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))
        self._coefA = CastedObjectData[0]

        """Coefficient B"""
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(2)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))
        self._coefB = CastedObjectData[0]

        """Coefficient C"""
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(3)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))
        self._coefC = CastedObjectData[0]

        """Coefficient D"""
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(4)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))
        self._coefD = CastedObjectData[0]

        """
        print('coefficients are %s %s %s %s' % (self._coefA, self._coefB, self._coefC, self._coefD))
        This gives the coefficients in some weird form, they're not what you expect them to be
        """

        self._doubleA = self._u32todouble(self._coefA)
        self._doubleB = self._u32todouble(self._coefB)
        self._doubleC = self._u32todouble(self._coefC)
        firstHalf = np.int16(self._coefD >> 16)
        secondHalf = np.int16(self._coefD & 0xffff)
        # Set the minimum and maximum wavelengths for the motor
        self._minwl = float(firstHalf)/10.0
        self._maxwl = float(secondHalf)/10.0
        # print 'first %s second %s' % (firstHalf, secondHalf)
        # This returns '10871' and '11859' for the Sacher, which are the correct
        # wavelength ranges in Angstroms
        #print 'Now calculate the current wavelength position:'
        self._currentwl = self._doubleA*(self._offset)**2.0 + self._doubleB*self._offset + self._doubleC
        #print('Current wavelength: %.3f nm' % self._currentwl)
        print('initializing done')
        print("")
        return True



#4)
    def get_offset(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL
        #DC - These are hardcoded values I got from the LabVIEW program -- I don't think any documentation exists on particular object indices
        StoredPositionObject = ctypes.wintypes.WORD(8321)
        StoredPositionObjectSubindex = ctypes.c_uint8(0)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_int32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_int32))
        if ret == 0:
            logging.error(__name__ + ' Could not read stored position from Sacher EPOS motor')
        print('motor offset value is: %s' % CastedObjectData[0])
        return CastedObjectData[0]


#5)
    def get_motor_position(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        pPosition = ctypes.pointer(ctypes.c_long())
        eposlib.VCS_GetPositionIs.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.c_long), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetPositionIs.restype = ctypes.wintypes.BOOL
        ret = eposlib.VCS_GetPositionIs(self._keyhandle, nodeID, pPosition, ctypes.byref(buf))
        print('motor position value is: %s' % pPosition.contents.value)
        return pPosition.contents.value

#6)
    def do_get_wavelength(self):
        self._offset = self.get_offset()
        self._motor_position = self.get_motor_position()
        self._currentwl1 = self._doubleA*(self._offset)**2.0 + self._doubleB*self._offset + self._doubleC
        self._currentwl2 = self._doubleA*(self._motor_position)**2.0 + self._doubleB*self._motor_position + self._doubleC
        print('Current wavelength according to offset: %.3f nm' % self._currentwl1)
        print('Current wavelength according to motor position: %.3f nm' % self._currentwl2)
        return self._currentwl1

    """
    And now we move on to setting things
    """

#7)
    def set_new_offset(self, new_offset):
        """
        This is NOT using the function "eposlib.VCS_MoveToPosition"
        So there's no literal motor movement is going on here
        It's just storing a value in some sort of saved memory on the instrument itself
        """
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        eposlib.VCS_SetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_SetObject.restype = ctypes.wintypes.BOOL
        StoredPositionObject = ctypes.wintypes.WORD(8321)
        StoredPositionObjectSubindex = ctypes.c_uint8(0)
        StoredPositionNbBytesToWrite = ctypes.wintypes.DWORD(4)
        ObjectDataArray = (ctypes.c_uint32*1)(new_offset)
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesWritten = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_SetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToWrite, StoredPositionNbBytesWritten, ctypes.byref(buf))
        if ret == 0:
            logging.error(__name__ + ' Could not write stored position from Sacher EPOS motor')
        return

#8)
    def set_coeffs(self, a, b, c, min_wl, max_wl):
        print('')
        print("setting coefficients...")
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        eposlib.VCS_SetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_SetObject.restype = ctypes.wintypes.BOOL
        d = (min_wl << 16) + max_wl
        StoredPositionObject = ctypes.wintypes.WORD(8204)

        for subidx, coeff in enumerate([a, b, c]):
            print(subidx, coeff)
            StoredPositionObjectSubindex = ctypes.c_uint8(subidx + 1)
            StoredPositionNbBytesToWrite = ctypes.wintypes.DWORD(4)
            ObjectDataArray = (ctypes.c_uint32*1)(self._doubletou32(coeff))
            ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
            StoredPositionNbBytesWritten = ctypes.pointer(ctypes.wintypes.DWORD(0))
            ret = eposlib.VCS_SetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToWrite, StoredPositionNbBytesWritten, ctypes.byref(buf))

        StoredPositionObjectSubindex = ctypes.c_uint8(4)
        StoredPositionNbBytesToWrite = ctypes.wintypes.DWORD(4)
        ObjectDataArray = (ctypes.c_uint32*1)(d)
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesWritten = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_SetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToWrite, StoredPositionNbBytesWritten, ctypes.byref(buf))

        print('Coefficients are %s %s %s' % (self._doubleA, self._doubleB, self._doubleC))

        if ret == 0:
            logging.error(__name__ + ' Could not write stored position from Sacher EPOS motor')
        return


#9)
    def set_target_position(self, target, absolute, immediately):
        """
        This is the function that actually moves the motor
        Since the motor position this thing reads can't be trusted, we're only doing relative movements and not absolute ones
        This means the "absolute" argument should always be set as "false"

        "target" is the target motor position, but be very careful what you want to set this to depending on whether you're doing an absolute or relative movement

        In the "absolute" category,
        True starts an absolute movement, False starts a relative movement
        Since the absolute position can't be trusted at all, we should always do "False", the relative movement

        In the "immediately" category,
        True starts immediately, False waits to end of last positioning
        We typically do True for this without any real issues

        The actual money function is "eposlib.VCS_MoveToPosition"
        """

        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_SetEnableState(self._keyhandle,nodeID,ctypes.byref(buf))
        pTarget = ctypes.c_long(target)
        pAbsolute = ctypes.wintypes.BOOL(absolute)
        pImmediately = ctypes.wintypes.BOOL(immediately)
        eposlib.VCS_MoveToPosition.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.c_long, ctypes.wintypes.BOOL, ctypes.wintypes.BOOL, ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_MoveToPosition.restype = ctypes.wintypes.BOOL
        ret = eposlib.VCS_MoveToPosition(self._keyhandle, nodeID, pTarget, pAbsolute, pImmediately, ctypes.byref(buf))
        steps_per_second = 14494.0 # hardcoded, estimated roughly, unused now
        nchecks = 0
        while nchecks < 1000:
            self._motor_position = self.get_motor_position()
            self._offset = self.get_offset()
            pMovementState = ctypes.pointer(ctypes.wintypes.BOOL())
            eposlib.VCS_GetMovementState.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.wintypes.BOOL), ctypes.POINTER(ctypes.wintypes.DWORD)]
            eposlib.VCS_GetMovementState.restype = ctypes.wintypes.BOOL
            ret = eposlib.VCS_GetMovementState(self._keyhandle, nodeID, pMovementState, ctypes.byref(buf))
            if pMovementState.contents.value == 1:
                break
            nchecks = nchecks + 1
            time.sleep(0.01)
        ret = eposlib.VCS_SetDisableState(self._keyhandle,nodeID,ctypes.byref(buf))
        return ret


#10)
    def do_set_wavelength(self, wavelength):
        """
        Here's the basic procedure:
        1) Convert the desired target wavelength into a motor position, keeping in mind that we're using the offset as the motor position
        x is what the motor position should be
        2) Calculate difference between the target position and the stored offset
        3) Prompt some confirmation before we crash this plane
        4) Then actually move the motor
        """
        current_wavelength = self.do_get_wavelength()
        current_offset = self.get_offset()

        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        x = (-1.0*self._doubleB + np.sqrt(self._doubleB**2.0 - 4.0*self._doubleA*(self._doubleC - wavelength))) / (2.0*self._doubleA)
        wavelength_to_pos = int(round(x))
        diff_wavelength_offset = wavelength_to_pos - int(self._offset)

        print('')
        print("The current wavelength, as according to the stored offset, is: %s" % current_wavelength)
        print("You're about to set the wavelength to: %s" % wavelength)
        print("This means moving the motor by %s steps" % diff_wavelength_offset)
        print("Where currently, the stored offset is: %s" % current_offset)
        # confirm = str(input("Is this ok? (y/n)"))
        #
        # if confirm == 'y':
        #     print('Ok then, setting wavelength...')
        #     print('')
        # else:
        #     print("ok then, shutting the whole thing down!")
        #     sys.exit(-1)

        if self._HPM and diff_wavelength_offset < 0:
            self.set_target_position(diff_wavelength_offset - 10000, False, True)
            self.set_target_position(10000, False, True)
        else:
            self.set_target_position(diff_wavelength_offset, False, True)

        self.set_new_offset(current_offset + diff_wavelength_offset)
        current_offset = self.get_offset()
        print("Now the stored offset is: %s" % current_offset)

        return

















































#11)
    @staticmethod
    def get_bit(byteval, idx):
    # def get_bit(self, byteval,idx):
        return ((byteval&(1<< idx ))!=0)
    """
    We take the two input numbers "byteval" and "idx" and do bitwise operations with them
    but it's not clear what the purpose of these bitwise operations are
    I also don't know what the @staticmethod decorator is supposed to be doing
    """

#12)
    @staticmethod
    def _u32todouble(uinput):
    # def _u32todouble(self, uinput):

        # get sign of number
        sign = Sacher_EPOS.get_bit(uinput,31)
        if sign == False:
            mantissa_sign = 1
        elif sign == True:
            mantissa_sign = -1
        exp_mask =  0b111111
        #print 'uin u is %d' % uinput
        #print 'type uin %s' % type(uinput)
        #print 'binary input is %s' % bin(long(uinput))
        # get sign of exponent
        if Sacher_EPOS.get_bit(uinput,7) == False:
            exp_sign = 1
        elif Sacher_EPOS.get_bit(uinput,7) == True:
            exp_sign = -1

        #print 'exp extract %s' % bin(int(uinput & exp_mask))
        #print 'exp conv %s' % (exp_sign*int(uinput & exp_mask))
        #print 'sign of exponent %s' % self.get_bit(uinput,7)
        #print 'binary constant is %s' % bin(int(0b10000000000000000000000000000000))
        mantissa_mask = 0b01111111111111111111111100000000
        # mantissa_mask = 0b0111111111111111111111110000000


        #print 'mantissa extract is %s' % bin((uinput & mantissa_mask) >> 8)
        mantissa = 1.0/1000000.0*float(mantissa_sign)*float((uinput & mantissa_mask) >> 8)
        #print 'mantissa is %.12f' % mantissa
        # print(1 if Sacher_EPOS.get_bit(uinput,31) else 0, mantissa, 1 if Sacher_EPOS.get_bit(uinput,7) else 0, uinput & exp_mask)
        output = mantissa*2.0**(float(exp_sign)*float(int(uinput & exp_mask)))
        #print 'output is %s' % output
        return output
    """
    ok dc gave some slight explanations here
    This function implements the "really weird/non-standard U32 to floating point conversion in the sacher VIs"
    It'd be gr8 if I knew what U32's were
    unsigned 32 bit something something?
    ah whatever
    Also I'm seeing mantissas and masks, this is bad
    """

#13)
    @staticmethod
    def _doubletou32(dinput):
        mantissa_bit = 0 if int(dinput / abs(dinput)) > 0 else 1
        exp_bit = 1 if -1 < dinput < 1 else 0

        b = np.ceil(np.log10(abs(dinput)))
        a = dinput / 10 ** b
        if dinput < 0:
            a = -a
        # print('a:\t{}\tb:\t{}'.format(a, b))

        d = np.log2(10) * b
        d_ = np.ceil(d)
        c = a * 2 ** (d - d_)
        # print('c:\t{}\td_:{}\toriginal:\t{}'.format(c, d_, c * 2 ** d_))

        return (int(mantissa_bit) << 31) + (int(c * 1e6) << 8) + (int(exp_bit) << 7) + int(abs(d_))

    """
    I think this was a new function that we made

    There might be a labview VI that does this correctly
    """

#14)
    def __del__(self):
        # execute disconnect
        self.close()
        return
    """
    this might be the only self explanatory one
    it disconnects
    """

#15)
    def close(self):
        print('closing EPOS motor.')

        eposlib.VCS_CloseDevice.argtypes = [ctypes.wintypes.HANDLE, ctypes.POINTER(DWORD)]
        eposlib.VCS_CloseDevice.restype = ctypes.wintypes.BOOL
        buf = ctypes.pointer(DWORD(0))
        ret = ctypes.wintypes.BOOL()

        ret = eposlib.VCS_CloseDevice(self._keyhandle, buf)

        #print 'close device returned %s' % buf

        if int(buf.contents.value) >= 0:
            self._is_open = False
        else:
            logging.error(__name__ + ' did not close Sacher EPOS motor correctly.')
        return
    """
    Apparently this closes the EPOS motor
    I don't know what "opening" and "closing" the motor means though
    and yeah also these random variables don't make any sense to me
    """


#16)
    def get_motor_current(self):
        nodeID = ctypes.wintypes.WORD(0)
        eposlib.VCS_GetCurrentIs.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetCurrentIs.restype = ctypes.wintypes.BOOL

        motorCurrent = ctypes.c_uint8(0)
        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_GetCurrentIs(self._keyhandle, nodeID, ctypes.byref(motorCurrent), ctypes.byref(buf))
        return motorCurrent.value

    """
    Not sure what this is doing yet
    """


#17)
    def find_home(self):
        nodeID = ctypes.wintypes.WORD(0)
        eposlib.VCS_FindHome.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_FindHome.restype = ctypes.wintypes.BOOL

        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_FindHome(self._keyhandle, nodeID, ctypes.c_uint8(35), ctypes.byref(buf))
        print('Homing: {}'.format(ret))
        return ret

    """
    Not sure what this is doing yet
    """


#18)
    def restore(self):
        nodeID = ctypes.wintypes.WORD(0)
        eposlib.VCS_FindHome.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_FindHome.restype = ctypes.wintypes.BOOL
        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_Restore(self._keyhandle, nodeID, ctypes.byref(buf))
        print('Restore: {}'.format(ret))
        return ret

    """
    Not sure what this is doing yet
    """



#19)
    def fine_tuning_steps(self, steps):
        current_motor_pos = self.get_motor_position()
        self._offset = self.get_offset()
        self.set_target_position(steps, False, True)
        new_motor_pos = self.get_motor_position()
        #print('New motor position is %s' % new_motor_pos)
        #print 'new offset is %s' % (new_motor_pos-current_motor_pos+self._offset)
        self.set_new_offset(new_motor_pos-current_motor_pos+self._offset)

    """
    Not sure what this is doing yet
    """

#20)
    def is_open(self):
        return self._is_open

#21)
    def clear_fault(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_ClearFault(self._keyhandle,nodeID,ctypes.byref(buf))
        print('clear fault buf %s, ret %s' % (buf, ret))
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            eposlib.VCS_GetErrorInfo(buf, errbuf, WORD(64))
            raise ValueError(errbuf.value)
    """
    Not sure what this is doing yet
    """




"""
We're done with the Sacher_EPOS() class at this point
"""



# -*- coding: utf-8 -*-

import os

from lantz import Feat
from lantz.foreign import LibraryDriver
from ctypes import c_char_p
from time import sleep

class FF(LibraryDriver):

    lib_path = os.path.join(os.environ['PROGRAMFILES'], 'Thorlabs\\Kinesis')
    lib_name = 'Thorlabs.MotionControl.FilterFlipper.dll'


    LIBRARY_NAME = os.path.join(lib_path, lib_name)
    LIBRARY_PREFIX = ''

    COM_DELAY = 0.2

    def __init__(self, serial_no, *args, **kwargs):
        """
        serial_no: unique device identifier; found on the label of the device
        """
        self.prev_path = os.environ['PATH']
        os.environ['PATH'] = '{}{}{}'.format(self.lib_path, os.pathsep, self.prev_path)
        super(FF, self).__init__(*args, **kwargs)
        os.environ['PATH'] = self.prev_path
        self.serial_no = c_char_p(str(serial_no).encode('ascii'))
        retval = self.lib.FF_Open(self.serial_no)
        if retval:
            raise RuntimeError('Could not initialize device: error {}'.format(retval))
        return

    @Feat()
    def position(self):
        return self.lib.FF_GetPosition(self.serial_no)

    @position.setter
    def position(self, value):
        self.lib.FF_MoveToPosition(self.serial_no, value)
        return

if __name__ == '__main__':
    import logging
    import sys
    from lantz.log import log_to_screen
    import numpy as np
    log_to_screen(logging.CRITICAL)
    serial_no = sys.argv[1]
    with FF(serial_no) as inst:
        print(inst.position)
        inst.position = 1
        print(inst.position)
        sleep(1)
        inst.position = 2



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.prior.nanoscanz
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Action, Feat
from lantz.drivers.legacy.serial import SerialDriver
from lantz import errors

class NanoScanZ(SerialDriver):
    """Driver for the NanoScanZ Nano Focusing Piezo Stage from Prior.
    """

    ENCODING = 'ascii'
    
    RECV_TERMINATION = '\r'
    SEND_TERMINATION = '\r'
    
    BAUDRATE = 9600
    BYTESIZE = 8
    PARITY = 'none'
    STOPBITS = 1

    #: flow control flags
    RTSCTS = False
    DSRDTR = False
    XONXOFF = False

    
    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        """Send query to the stage and return the answer, after handling
        possible errors.

        :param command: command to be sent to the instrument
        :type command: string

        :param send_args: (termination, encoding) to override class defaults
        :param recv_args: (termination, encoding) to override class defaults
        """
        ans = super().query(command, send_args=send_args, recv_args=recv_args)
        if ans[0] == 'E':
            code = ans[2]
            if code == '8':
                raise errors.InstrumentError('Value out of range')
            elif code == '4':
                raise errors.InstrumentError('Command parse error, ie wrong number  of parameters')
            elif code == '5':
                raise errors.InstrumentError('Unknown command')
            elif code == '2':
                raise errors.InstrumentError('Invalid checksum')

        return ans
    
    @Feat(values={9600,19200,38400})
    def baudrate(self):
        """Reports and sets the baud rate.
        NOTE: DO NOT change the baud rate of the Piezo controller when daisy chained to ProScan.
        """
        return self.query('BAUD')

    @baudrate.setter
    def baudrate(self, value):
        self.query('BAUD {}'.format(value))


    @Feat(values={True: '4', False: '00000'})
    def moving(self):
        """Returns the movement status, 0 stationary, 4 moving
        """
        return self.query('$')
    
    @Feat(read_once=True)
    def idn(self):
        """Identification of the device
        """
        return self.query('DATE') +' '+ self.query('SERIAL')
    
    @Feat(units = 'micrometer')
    def position(self):
        """Gets and sets current position.
        If the value is set to z = 0, the display changes to REL 0 (relative display mode). To return to ABS mode use inst.move_absolute(0) and then inst.position = 0. Thus, the stage will return to 0 micrometers and the display screen will switch to ABS mode.
        """
        return self.query('PZ')

    @position.setter
    def position(self, value):
        self.query('PZ {}'.format(value))
    
    @Action()
    def zero_position(self):
        """Move to zero including any position redefinitions done by the position Feat
        """
        self.query('M')
    
    
    @Action(units = 'micrometer', limits=(100,))
    def move_absolute(self, value):
        """Move to absolute position n, range (0,100).
        This is a "real" absolute position and is independent of any relative offset added by the position Feat.
        """
        self.query('V {}'.format())
    
    
    @Action()
    def move_relative(self, value):
        """ 
        Move the stage position relative to the current position by an amount determined by 'value'.
        If value is given in micrometer, thats the amount the stage is going to move, in microns.
        If value is given in steps, the stage will move a distance  value.magnitude * step. The step is defined by the step Feat
        """
        """
    try:
            u = value.units
            if value.magnitude > 0:
                self.query('U {}'.format(value.magnitude))
            if value.magnitude < 0:
                self.query('D {}'.format(-value.magnitude))
        except:
            if isinstance(value, int):
                if value > 0:
                    for x in range(0,value):
                        self.query('U')
                elif value < 0:
                    for x in range(0,value):
                        self.query('D')
            else:
                raise ValueError('Specify the translation distance in micrometer unit')               
        """
        try:
            if value.units == 'micrometer':
                if value.magnitude > 0:
                    self.query('U {}'.format(value.magnitude))
                elif value.magnitude < 0:
                    self.query('D {}'.format(-value.magnitude))
            elif value.units == 'steps':
                if value.magnitude > 0:
                    for x in range(0,value.magnitude):
                        self.query('U')
                elif value.magnitude < 0:
                    for x in range(0,-value.magnitude):
                        self.query('D')
        except:
            raise ValueError('Specify the translation distance in micrometers or steps') 
    

    @Feat(units='micrometer')
    def step(self):
        """Report and set the default step size, in microns
        """
        return self.query('C')

    @step.setter
    def step (self, value):
        self.query('C {}'.format(value))
    
    
    
    @Feat(read_once=True)
    def software_version(self):
        """Software version
        """  
        return self.query('VER')

class NanoScanZ_chained(NanoScanZ):
    """This is needed when the NanoScanZ controller is connected to a ProScanII controller through its RS-232-2 input
    """
    def write(self, command, termination=None, encoding=None):
        super().write('<' + command, termination=termination, encoding=encoding)

    def read(self, termination=None, encoding=None, recv_chunk=None):
        return super().read(termination=termination, encoding=encoding)[1:]


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Prior NanoScanZ')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_screen(lantz.log.DEBUG)
    with NanoScanZ(args.port) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            from lantz import Q_
            # Add your test code here
            print('Non interactive mode')
            
            um = Q_(1, 'micrometer')
            
            print(inst.idn)
            print(inst.moving)
            print(inst.position)
            print(inst.step)
            inst.step = 1 * um
            print(inst.step)



# -*- coding: utf-8 -*-
"""
    lantz.drivers.newport.fsm300
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of FSM300 using NI DAQ controller

    Author: Kevin Miao
    Date: 9/27/2016
"""

from lantz import Driver
from lantz.driver import Feat, DictFeat, Action
from lantz.drivers.ni.daqmx import AnalogOutputTask, VoltageOutputChannel

from lantz import Q_

import time

import numpy as np

def enforce_units(val, units):
    if not isinstance(val, Q_):
        val = Q_(val, units)
    else:
        val = val.to(units)
    return val

def enforce_point_units(point, units='um'):
    point = enforce_units(point[0], units) , enforce_units(point[1], units)
    return point


class FSM300(Driver):

    def __init__(self, x_ao_ch, y_ao_ch,
                 ao_smooth_rate=Q_('10 kHz'), ao_smooth_steps=Q_('1000 1/V'),
                 limits=((Q_(-10, 'V'), Q_(10, 'V')), (Q_(-10, 'V'), Q_(10, 'V'))),
                 cal=(Q_(10, 'um/V'), Q_(7.6, 'um/V'))):
        self.x_limits_mag = tuple(float(val / Q_('1 V')) for val in limits[0])
        self.y_limits_mag = tuple(float(val / Q_('1 V')) for val in limits[1])

        self.ao_smooth_rate = ao_smooth_rate
        self.ao_smooth_steps = ao_smooth_steps
        self.cal = cal

        self._position = (Q_('0 um'), Q_('0 um'))

        super().__init__()

        self.x_ao_ch = x_ao_ch
        self.y_ao_ch = y_ao_ch

        return

    def initialize(self):
        self.task = AnalogOutputTask('fsm300')
        VoltageOutputChannel(self.x_ao_ch, name='fsm_x', min_max=self.x_limits_mag, units='volts', task=self.task)
        VoltageOutputChannel(self.y_ao_ch, name='fsm_y', min_max=self.y_limits_mag, units='volts', task=self.task)
        super().initialize()

    def finalize(self):
        self.task.clear()
        super().finalize()


    def ao_smooth_func(self, init_point, final_point):
        init_x, init_y = init_point
        final_x, final_y = final_point

        init_x_voltage, final_x_voltage = init_x / self.cal[0], final_x / self.cal[0]
        init_y_voltage, final_y_voltage = init_y / self.cal[1], final_y / self.cal[1]
        diff_x_voltage = final_x_voltage - init_x_voltage
        diff_y_voltage = final_y_voltage - init_y_voltage

        diff_voltage = max(abs(diff_x_voltage), abs(diff_y_voltage))
        steps = int(np.ceil(diff_voltage * self.ao_smooth_steps))
        init = np.array([val.to('V').magnitude for val in [init_x_voltage, init_y_voltage]])
        diff = np.array([val.to('V').magnitude for val in [diff_x_voltage, diff_y_voltage]])

        versine_steps = (1.0 - np.cos(np.linspace(0.0, np.pi, steps))) / 2.0

        step_voltages = np.outer(np.ones(steps), init) + np.outer(versine_steps, diff)
        return step_voltages

    def ao_linear_func(self, init_point, final_point, steps):
        init_x, init_y = init_point
        final_x, final_y = final_point

        init_x_voltage, final_x_voltage = init_x / self.cal[0], final_x / self.cal[0]
        init_y_voltage, final_y_voltage = init_y / self.cal[1], final_y / self.cal[1]
        diff_x_voltage = final_x_voltage - init_x_voltage
        diff_y_voltage = final_y_voltage - init_y_voltage

        diff_voltage = max(abs(diff_x_voltage), abs(diff_y_voltage))
        init = np.array([val.to('V').magnitude for val in [init_x_voltage, init_y_voltage]])
        diff = np.array([val.to('V').magnitude for val in [diff_x_voltage, diff_y_voltage]])

        linear_steps = np.linspace(0.0, 1.0, steps)

        step_voltages = np.outer(np.ones(steps), init) + np.outer(linear_steps, diff)
        return step_voltages

    @Feat()
    def abs_position(self):
        return self._position


    @abs_position.setter
    def abs_position(self, point):
        point = enforce_point_units(point)
        step_voltages = self.ao_smooth_func(self._position, point)
        if step_voltages.size:
            steps = step_voltages.shape[0]
            clock_config = {
                'source': 'OnboardClock',
                'rate': self.ao_smooth_rate.to('Hz').magnitude,
                'sample_mode': 'finite',
                'samples_per_channel': steps,
            }
            self.task.configure_timing_sample_clock(**clock_config)
            self.task.configure_trigger_disable_start()
            task_config = {
                'data': step_voltages,
                'auto_start': False,
                'timeout': Q_(0,'s'),
                'group_by': 'scan',
            }
            self.task.write(**task_config)
            self.task.start()
            time.sleep((steps / self.ao_smooth_rate).to('s').magnitude)
            self.task.stop()
        self._position = point

    @Action()
    def line_scan(self, init_point, final_point, steps, acq_task, acq_rate=Q_('20 kHz'), pts_per_pos=100):
        init_point = enforce_point_units(init_point)
        final_point = enforce_point_units(final_point)
        timeout = enforce_units(1.5*(pts_per_pos*steps/acq_rate), units='s')
        # AO smooth move to initial point
        self.abs_position = init_point
        step_voltages = self.ao_linear_func(init_point, final_point, steps)
        if acq_task.IO_TYPE == 'CI':
            chs = list(acq_task.channels.keys())
            if not chs:
                raise ValueError('acquisition task must have at least one channel')
            ch = chs[0]
            dev = ch.split('/')[0]

            step_voltages = np.repeat(step_voltages, pts_per_pos + 1, axis=0)
            clock_config = {
                'rate': acq_rate.to('Hz').magnitude,
                'sample_mode': 'finite',
                'samples_per_channel': len(step_voltages),
            }
            self.task.configure_timing_sample_clock(**clock_config)
            clock_config = {
                'source': '/{}/ao/SampleClock'.format(dev),
                'rate': acq_rate.to('Hz').magnitude,
                'sample_mode': 'finite',
                'samples_per_channel': len(step_voltages),
            }
            acq_task.configure_timing_sample_clock(**clock_config)
            task_config = {
                'data': step_voltages,
                'auto_start': False,
                'timeout': Q_('0 s'),
                'group_by': 'scan',
            }
            self.task.write(**task_config)
            acq_task.arm_start_trigger_source = 'ao/StartTrigger'.format(dev)
            acq_task.arm_start_trigger_type = 'digital_edge'
            acq_task.start()
            self.task.start()
            scanned = acq_task.read(samples_per_channel=len(step_voltages), timeout=timeout)
            acq_task.stop()
            self.task.stop()
            scanned = scanned.reshape((steps, pts_per_pos + 1))
            averaged = np.diff(scanned).mean(axis=1)
            return averaged*acq_rate.to('Hz').magnitude
        elif acq_task.IO_TYPE == 'AI':
            step_voltages = np.repeat(step_voltages, pts_per_pos, axis=0)
            clock_config = {
                'source': 'OnboardClock',
                'rate': acq_rate.to('Hz').magnitude,
                'sample_mode': 'finite',
                'samples_per_channel': len(step_voltages),
            }
            self.task.configure_timing_sample_clock(**clock_config)
            acq_task.configure_timing_sample_clock(**clock_config)
            task_config = {
                'data': step_voltages,
                'auto_start': False,
                'timeout': Q_('0 s'),
                'group_by': 'scan',
            }
            self.task.write(**task_config)
            self.task.configure_trigger_digital_edge_start('ai/StartTrigger')
            self.task.start()
            acq_task.start()
            scanned = acq_task.read(samples_per_channel=len(step_voltages), timeout=timeout)
            acq_task.stop()
            self.task.stop()
            scanned = scanned.reshape((steps, pts_per_pos))
            averaged = scanned.mean(axis=1)
            return averaged
        else:
            pass



# -*- coding: utf-8 -*-
"""
    lantz.drivers.princetoninstruments.winspec
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Implementation of Winspec PI Camera over a Socket
    Author: Berk Diler
    Date: 29/08/2017
"""


from .winspec import Winspec

__all__ = ['Winspec']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.labjack
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :company: LabJack
    :description: LabJacks are USB/Ethernet based measurement and automation 
    devices which provide analog inputs/outputs, digital inputs/outputs, and more. 
    They serve as an inexpensive and easy to use interface between computers and
    the physical world.
    :website: http://www.labjack.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .u12 import U12

__all__ = ['U12']




# -*- coding: utf-8 -*-
"""
    lantz.drivers.aeroflex
    ~~~~~~~~~~~~~~~~~~~~~~

    :company: Aeroflex
    :description: Test and measurement equipment and microelectronic solutions.
    :website: http://www.aeroflex.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .a2023a import A2023a

__all__ = ['A2023a', ]




# -*- coding: utf-8 -*-
"""
    lantz.drivers.oceanoptics.usb4000
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Driver for spectrograph.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


import struct
import numpy as np

from pyvisa.errors import VisaIOError

from lantz import Feat, DictFeat
from lantz.drivers.legacy.usb import USBDriver, usb_find_desc

__all__ = ['USB4000']

commands = {
    0x01: ('init', 'initialize USB4000'), 
    0x02: ('set_i', 'set integration time in uS'), 
    0x03: ('set_strobe', 'set strobe enable status'), 
    0x05: ('ask', 'query information'), 
    0x06: ('write', 'write information'), 
    0x09: ('get_spectra', 'request spectra'),
    0x0A: ('set_trigger', 'set trigger mode'),
    0x0B: ('num_plugins', 'query number of plug-in accessories present'),
    0x0C: ('plugin_ids', 'query plug-in identifiers'),
    0x0D: ('detect_plugins', 'detect plug-ins'),
    0x60: ('read_i2c', 'general I2C read'),
    0x61: ('write_i2c', 'general I2C write'),
    0x62: ('spi_io', 'general spi I/O'),
    0x68: ('read_psoc', 'PSOC read'),
    0x69: ('write_psoc', 'PSOC write'),
    0x6A: ('write_reg', 'write register information'),
    0x6B: ('read_reg', 'read register information'),
    0x6C: ('read_temp', 'read PCB temperature'),
    0x6D: ('read_calib', 'read irradiance calibration factors'),
    0x6E: ('write_calib', 'write irradiance calibration factors'),
    0xFE: ('ask_2', 'query information')
}

config_regs = {
    0: 'serial_number',
    1: '0_order_wavelength_coeff',
    2: '1_order_wavelength_coeff',
    3: '2_order_wavelength_coeff',
    4: '3_order_wavelength_coeff',
    5: 'stray_light_constant',
    6: '0_order_nonlinear_coeff',
    7: '1_order_nonlinear_coeff',
    8: '2_order_nonlinear_coeff',
    9: '3_order_nonlinear_coeff',
    10: '4_order_nonlinear_coeff',
    11: '5_order_nonlinear_coeff',
    12: '6_order_nonlinear_coeff',
    13: '7_order_nonlinear_coeff',
    14: 'polynomial_order',
    15: 'bench_configuration',
    16: 'USB4000_config',
    17: 'autonull',
    18: 'baud_rate'
}   


class USB4000(USBDriver):
    """Ocean Optics spectrometer
    """

    def __init__(self, serial_number=None, **kwargs):
        super().__init__(vendor=0x2457, product=0x1022, serial_number=serial_number, **kwargs)

        self._spec_hi = usb_find_desc(self.usb_inf, bEndpointAddress=0x82)
        self._spec_lo = usb_find_desc(self.usb_inf, bEndpointAddress=0x86)

        # initialize spectrometer
        self.initialize()

    def initialize(self):
        """Ends the initialization command to the USB4000 spectrometer
            
        This command initializes certain parameters on the USB4000 and sets internal variables
        based on the USB communication speed. This command is called at object instantiation.
        """
        self.usb_send_ep.write(struct.pack('<B', 0x01))

    integration_time = Feat(units='microsecond', limits=(10, 65535000))

    @integration_time.setter
    def integration_time(self, dt):
        """Sets the integration time to use by the spectrometer in microseconds

        This command sets the integration time that the USB4000 uses when acquiring spectra. The
        value is passed as a 32 bit integer and has a valid range between 10 and 65,535,000 us. 
        """

        cmd = struct.pack('<BI', 0x02, dt)
        self.usb_send_ep.write(cmd)

    @DictFeat(keys=tuple(range(32)))
    def query_config(self, position):
        """returns the value stored in register `reg` of the spectrometer
        """
        cmd = struct.pack('<2B', 0x05, position)
        self.usb_send_ep.write(cmd)
        resp = bytearray(self.usb_recv_ep.read(64, timeout=1000))

        # Check that proper echo is returned
        assert resp[0:2] == bytearray([0x05, position]) 

        config.update({config_regs[x] : resp[2:].decode('ascii', errors='ignore').strip('\x00')})
            
        return config
            
    def reset(self):
        self.usb.reset()

    @Feat(units='degree_celsius')
    def pcb_temperature(self):
        cmd = struct.pack('<B', 0x6C)
        self.usb_send_ep.write(cmd)        
        resp = bytearray(self.usb_recv_ep.read(3, timeout=1000))
        
        # Check that proper echo is returned
        assert resp[0:1] == bytearray([0x08]) 
        
        t = struct.unpack('<h', resp[1:3])[0]
        return t * 0.003906
    
    @Feat(read_once=True)
    def firmware_version(self):
        cmd = struct.pack('<2B', 0x6B, 0x04)
        self.usb_send_ep.write(cmd)
        
        resp = bytearray(self.usb_recv_ep.read(3, timeout=200))
        log.debug('got {:s}'.format(repr(resp)))
        
        assert resp[0:1] == bytearray([0x04]) # Check that proper echo is returned
        
        vers = struct.unpack('>H', resp[1:3])[0]
        log.info('firmware is {:d}'.format(vers))
        return vers

    trigger = Feat(values={'Freerun': 0, 'Software': 1, 'ExternalSync': 2, 'ExternalHard': 3})

    @trigger.setter
    def trigger(self, mode):
        cmd = struct.pack('<BH', 0x0A, mode)
        self.usb_send_ep.write(cmd)

    def request_spectra(self):
        cmd = struct.pack('<B', 0x09)
        self.log_debug('Requesting spectra')
        self.usb_send_ep.write(cmd)
        
        data = np.zeros(shape=(3840,), dtype='<u2')
        
        try:
            data_lo = self._spec_lo.read(512*4, timeout=100)
            data_hi = self._spec_hi.read(512*11, timeout=100)

            data_sync = self._spec_hi.read(1, timeout=100)

            assert struct.unpack('<B', data_sync)[0] == 0x69

            data[:1024], data[1024:] = np.frombuffer(data_lo, dtype='<u2'), \
                                       np.frombuffer(data_hi, dtype='<u2')
        except AssertionError:
            self.log_error('Not synchronized')
        except VisaIOError:
            self.log_error('Timeout on usb')
        finally:
            self.log_debug('Obtained spectra')

        return data

    def get_status(self):
        cmd = struct.pack('<B', 0xFE)
        self.usb_send_ep.write(cmd)
        resp = bytearray(self.usb_recv_ep.read(16, timeout=1000))

        stat = {
            'num_pixels': struct.unpack('<H', resp[0:2])[0],
            'integration_time': struct.unpack('<I', resp[2:6])[0],
            'lamp_enable': bool(struct.unpack('<B', resp[6:7])[0]),
            'trigger_mode': struct.unpack('<B', resp[7:8])[0],
            'acq_status': struct.unpack('<B', resp[8:9])[0],
            'packets_in_spectra': struct.unpack('<B', resp[9:10])[0],
            'power_down': bool(struct.unpack('<B', resp[10:11])[0]),
            'packet_count': struct.unpack('<B', resp[11:12])[0],
            'usb_speed': struct.unpack('<B', resp[14:15])[0]
        }
        
        return stat



# -*- coding: utf-8 -*-
"""
    lantz.drivers.tektronix
    ~~~~~~~~~~~~~~~~~~~~~~~

    :company: Tektronix.
    :description: Test and Measurement Equipment.
    :website: http://www.tek.com/

    ---

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD,

"""

from .tds2024b import TDS2024
from .tds1012 import TDS1012
from .tds1002b import TDS1002b

__all__ = ['TDS2024', 'TDS1002b', 'TDS1012']

